#include "modelcanvas.h"
#include "video.h"
#include "animcontrol.h"
#include "gifexporter.h"

//float animSpeed = 1.0f;
const float piover180 = 0.0174532925f;
const float rad2deg = 57.295779513f;

DWORD lastTime = 0;

IMPLEMENT_CLASS(ModelCanvas, wxGLCanvas)

BEGIN_EVENT_TABLE(ModelCanvas, wxGLCanvas)
    EVT_SIZE(ModelCanvas::OnSize)
    EVT_PAINT(ModelCanvas::OnPaint)
    EVT_MOUSE_EVENTS(ModelCanvas::OnMouse)
	EVT_KEY_DOWN(ModelCanvas::OnKey)

    EVT_TIMER(ID_TIMER, ModelCanvas::OnTimer)
	//EVT_IDLE(ModelCanvas::OnIdle)

END_EVENT_TABLE()

#ifdef _WIN32
	// for timeGetTime:
	#pragma comment(lib,"Winmm.lib")

#else // for linux
	#include <sys/time.h>

	//typedef int DWORD;
	int timeGetTime()
	{
		static int start=0;
		static struct timeval t;
		gettimeofday(&t, NULL);
		if (start==0){
			start = t.tv_sec;
		}

		return (int)((t.tv_sec-start)*1000 + t.tv_usec/1000);
	}
#endif

DWORD pauseTime;

ModelCanvas::ModelCanvas(wxWindow *parent, wxWindowID id, const wxPoint& pos, const wxSize& size, long style, const wxString& name, int *attribList, const wxPalette& palette, const int pixelFormat)
    : wxGLCanvas(parent, id, pos, size, style|wxFULL_REPAINT_ON_RESIZE, name, attribList, palette, pixelFormat)
{
    init = false;
	render = false;
	
	srand(timeGetTime());

	time = 0;
	lastTime = timeGetTime();

	timer.SetOwner(this, ID_TIMER);
	timer.Start(TIME_STEP);
	//wxIdleEvent::SetMode(wxIDLE_PROCESS_SPECIFIED);

	fov = 45.0f;
	vRot = Vec3D(0.0f, 0.0f, 0.0f);
	vPos = Vec3D(0.0f, 0.0f, 5.0f);

	lightType = LT_DIRECTIONAL;

	// Setup our default colour values.
	bgcolor = Vec3D(0.1f, 0.2f, 0.4f);
	baseTrans = Vec3D(0.0f, 0.0f, 0.0f);

	model =	0;			// Main model.
	skyModel = 0;		// SkyBox Model
	wmo = 0;			// world object model
	animControl = 0;
	viewControl = 0;
	gifExporter = 0;
	lightControl = 0;	
	lights = 0;			// Lights
	rt = 0;				// RenderTexture class
	curAtt = 0;			// Current Attachment

	//drawParticles = true;
	drawMask = false;
	drawLightDir = false;

	drawBackground = false;
	drawSky = false;
	bMouseLight = false;
	useCamera = false;
	
	root = new Attachment(NULL, NULL, -1, -1);
	sky = new Attachment(NULL, NULL, -1, -1);

	Show(true);
}

ModelCanvas::~ModelCanvas()
{
	clearAttachments();

	wxDELETE(root);
	wxDELETE(sky);
	//wxDELETE(wmo);
	//wxDELETE(model);

	//wxDELETEA(lights);

	if (rt) {
		rt->Shutdown();
		wxDELETE(rt);
	}
}


void ModelCanvas::OnSize(wxSizeEvent& event)
{
	wxGLCanvas::OnSize(event);

	if (init) 
		InitView();
	else
		InitGL();
}

void ModelCanvas::SetupProjection(float scale)
{
	glMatrixMode(GL_PROJECTION);
    glLoadIdentity();	
	if (scale < 100.0f) {
		// "near" projection
	    gluPerspective(fov, aspect, 0.1f, 128.0f);

	} else if (scale >= 100.0f && scale < 512.0f) {
		// "far" projection
	    gluPerspective(fov, aspect, 1.0f, 512.0f);

	} else {
		// very far projection
		gluPerspective(fov, aspect, 3.0f, 1536.f);
	}
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}


void ModelCanvas::InitView()
{
	// set GL viewport
	int w, h;
	GetClientSize(&w, &h);
	SetCurrent();
	glViewport(0, 0, (GLint) w, (GLint) h);

	aspect = (float)w / (float)h;

	gXres = w;
	gYres = h;

	// update projection
	SetupProjection();
}

Attachment* ModelCanvas::LoadModel(const char *fn)
{
	clearAttachments();
	root->model = NULL;

	wxDELETE(wmo);

	model = new Model(fn, true);
	if (!model->ok) {
		wxDELETE(model);
		model = NULL;
		return NULL;
	}
	root->model = model;

	curAtt = root;

	ResetView();
	SetupProjection(modelsize);
	
	baseTrans = Vec3D(0.0f, 0.0f, 0.0f);

	return root;
}

Attachment* ModelCanvas::AddModel(const char *fn)
{
	Attachment *att = root->addChild(fn, (int)root->children.size(), -1);
	
	curAtt->pos = vPos;
	curAtt->rot = vRot;

	curAtt = att;

	ResetView();
	SetupProjection(modelsize);

	return att;
}

Attachment* ModelCanvas::LoadCharModel(const char *fn)
{
	clearAttachments();
	root->model = NULL;

	wxDELETE(wmo);

	// Create new one
	model = new Model(fn, true);
	if (!model->ok) {
		wxDELETE(model);
		model = NULL;
		return NULL;
	}
	baseTrans = Vec3D(0.0f, 0.0f, 0.0f);
	ResetView();
	SetupProjection(modelsize);

	Attachment *att = root->addChild(model, 0, -1);
	curAtt = att;
	return att;
}

void ModelCanvas::LoadWMO(std::string fn)
{
	if (!wmo) {
		wmo = new WMO(fn);
		root->model = wmo;
	}
}


void ModelCanvas::clearAttachments()
{
	root->delChildren();
	sky->delChildren();
}

void ModelCanvas::OnMouse(wxMouseEvent& event)
{
	if (!model && !wmo)
		return;

	int px = event.GetX();
	int py = event.GetY();
	int pz = event.GetWheelRotation();

	// mul = multiplier in which to multiply everything to achieve a sense of control over the amount to move stuff by
	float mul = 1.0f;
	if (event.m_shiftDown)
		mul /= 10;
	if (event.m_controlDown)
		mul *= 10;
	if (event.m_altDown)
		mul *= 50;

	if (wmo) {

		//if (model->animManager)
		//	mul *= model->animManager->GetSpeed(); //animSpeed;

		if (event.ButtonDown()) {
			mx = px;
			my = py;
			vRot0 = vRot;
			vPos0 = vPos;

		} else if (event.Dragging()) {
			int dx = mx - px;
			int dy = my - py;

			if (event.LeftIsDown()) {
				vRot.x = vRot0.x - (dy / 2.0f); // * mul);
				vRot.y = vRot0.y - (dx / 2.0f); // * mul);

				//viewControl->Refresh();

			} else if (event.RightIsDown()) {
				mul /= 100.0f;

				vPos.x = vPos0.x - dx*mul;
				vPos.y = vPos0.y + dy*mul;

				//viewControl->Refresh();

			} else if (event.MiddleIsDown()) {
				mul = (mul / 20.0f) * dy;

				Zoom(mul, false);
				my = py;
			}

		} else if (event.GetEventType() == wxEVT_MOUSEWHEEL) {
			if (pz != 0) {
				mul = (mul / 120.0f) * pz;
				Zoom(mul, false);
			}
		}

	} else if (model) {
		if (model->animManager)
			mul *= model->animManager->GetSpeed(); //animSpeed;

		if (event.ButtonDown()) {
			mx = px;
			my = py;
			if (bMouseLight) // going to use vRot to hold our temp light position (technically, our lights rotation).
				vRot0 = Vec3D(lightControl->GetCurrentPos().x,lightControl->GetCurrentPos().y,lightControl->GetCurrentPos().z);
			else
				vRot0 = vRot;
			vPos0 = vPos;

		} else if (event.Dragging()) {
			int dx = mx - px;
			int dy = my - py;

			if (event.LeftIsDown()) {
				if (bMouseLight)
					return;

				vRot.x = vRot0.x - (dy / 2.0f); // * mul);
				vRot.y = vRot0.y - (dx / 2.0f); // * mul);

				viewControl->Refresh();

			} else if (event.RightIsDown()) {
				mul /= 100.0f;

				if (bMouseLight) {
					Vec4D temp = lightControl->GetCurrentPos();
					temp.y = vRot0.y + dy*mul;
					temp.x = vRot0.x - dx*mul;
					lightControl->SetPos(temp);
					lightControl->Update();
				} else {
					vPos.x = vPos0.x - dx*mul;
					vPos.y = vPos0.y + dy*mul;

					viewControl->Refresh();
				}

			} else if (event.MiddleIsDown()) {
				if (!event.m_altDown) {
					mul = (mul / 20.0f) * dy;

					if (bMouseLight) {
						Vec4D temp = lightControl->GetCurrentPos();
						temp.z = vRot0.z - mul;
						lightControl->SetPos(temp); 
						lightControl->Update();
					} else {
						Zoom(mul, false);
						my = py;
					}

				} else {
					mul = (mul / 1200.0f) * dy;
					Zoom(mul, true);
					my = py;
				}
			}

		} else if (event.GetEventType() == wxEVT_MOUSEWHEEL) {
			if (pz != 0) {
				mul = (mul / 120.0f) * pz;
				if (!wxGetKeyState(WXK_ALT)) {
					if (bMouseLight) {
						Vec4D temp = lightControl->GetCurrentPos();
						temp.z -= mul / 10.0f;
						lightControl->SetPos(temp); 
						lightControl->Update();
					} else {
						Zoom(mul, false);
					}
				} else {
					mul /= 50.0f;
					Zoom(mul, true);
				}
			}
		}
	}

	if (event.GetEventType() == wxEVT_ENTER_WINDOW)
		SetFocus();
}

void ModelCanvas::InitGL()
{
	SetCurrent();

	// init stuff
	InitExtensions();

	if (supportDrawRangeElements || supportVBO) {
		glEnableClientState(GL_VERTEX_ARRAY);
		glEnableClientState(GL_NORMAL_ARRAY);
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	}

	if(supportAntiAlias && useAntiAlias)
		glEnable(GL_MULTISAMPLE_ARB);

	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
	//glColorMaterial(GL_FRONT, GL_SPECULAR);
	
	/*
	//glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_COMBINE);
	glTexEnvf(GL_TEXTURE_ENV,GL_RGB_SCALE, 1.000000);
	glTexEnvi(GL_TEXTURE_ENV,GL_SOURCE0_RGB, GL_PREVIOUS);
	glTexEnvi(GL_TEXTURE_ENV,GL_SOURCE1_RGB, GL_TEXTURE);
	glTexEnvi(GL_TEXTURE_ENV,GL_SOURCE2_RGB, GL_CONSTANT);
	glTexEnvi(GL_TEXTURE_ENV,GL_COMBINE_RGB, GL_MODULATE);
	glTexEnvf(GL_TEXTURE_ENV,GL_ALPHA_SCALE, 1.000000);
	glTexEnvi(GL_TEXTURE_ENV,GL_SOURCE0_ALPHA, GL_PREVIOUS);
	glTexEnvi(GL_TEXTURE_ENV,GL_SOURCE1_ALPHA, GL_TEXTURE);
	glTexEnvi(GL_TEXTURE_ENV,GL_SOURCE2_ALPHA, GL_CONSTANT);
	glTexEnvi(GL_TEXTURE_ENV,GL_COMBINE_ALPHA, GL_REPLACE);
	*/

	glAlphaFunc(GL_GEQUAL, 0.5f);
	glDepthFunc(GL_LEQUAL);
	glMaterialf(GL_FRONT, GL_SHININESS, 18.0f);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glPixelStorei(GL_PACK_ALIGNMENT, 4);
	glPixelStorei(GL_PACK_SWAP_BYTES, false);
	glPixelStorei(GL_PACK_LSB_FIRST, false);

	glShadeModel(GL_SMOOTH);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

	InitView();
	init = true;

	// Setup lighting
	lightControl->Init();

	/*
	// test
	GLuint retcode;
	//char *program_string = "!!ARBfp1.0.PARAM c[2] = { { 1, 0.29899999, 0.58700001, 0.14399999 },...{ 4 } };.TEMP R0;.TEMP R1;.TEX R0.xyz, fragment.texcoord[1], texture[1], RECT;.TEX R1.xyz, fragment.texcoord[0], texture[0], RECT;.MUL R0.xyz, R0, R0;.MAD R0.xyz, R0, fragment.color.primary.w, R1;.DP3_SAT R0.x, R0, c[0].yzww;.ADD R0.y, -R0.x, c[0].x;.MUL R0.y, R0.x, R0;.MUL_SAT R0.y, R0, c[1].x;.MAD result.color.xyz, fragment.color.primary, R0.y, R0.x;.MOV result.color.w, c[0].x;.END"; //Death
	//char *program_string = "!!ARBfp1.0.PARAM c[2] = { { 1, 0.29899999, 0.58700001, 0.14399999 },...{ 4 } };.TEMP R0;.TEMP R1;.TEX R0.xyz, fragment.texcoord[1], texture[1], 2D;.TEX R1.xyz, fragment.texcoord[0], texture[0], 2D;.MUL R0.xyz, R0, R0;.MAD R0.xyz, R0, fragment.color.primary.w, R1;.DP3_SAT R0.x, R0, c[0].yzww;.ADD R0.y, -R0.x, c[0].x;.MUL R0.y, R0.x, R0;.MUL_SAT R0.y, R0, c[1].x;.MAD result.color.xyz, fragment.color.primary, R0.y, R0.x;.MOV result.color.w, c[0].x;.END"; //Death 2
	char *program_string = "!!ARBfp1.0.PARAM c[1] = { { 0.29899999, 0.58700001, 0.114 } };.TEMP R0;.TEX R0, fragment.texcoord[0], texture[0], 2D;.MUL result.color.w, fragment.color.primary, R0;.DP3 result.color.xyz, R0, c[0];.END"; // Desaturate
	glEnable(GL_FRAGMENT_PROGRAM_ARB);
	glGenProgramsARB(1, &retcode);
	glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, retcode);

	glProgramStringARB(GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, strlen(program_string), program_string);
	// -- */
}

struct Vertex
{
    float tu, tv;
    float x, y, z;
};

void ModelCanvas::OnPaint(wxPaintEvent& WXUNUSED(event))
{
	// Set this window handler as the reference to draw to.
	wxPaintDC dc(this);

	if (!init)
		InitGL();

	if (render) {

		if (wmo)
			RenderWMO();
		else 
			Render();
	}
}

inline void ModelCanvas::RenderLight(Light *l)
{
	GLUquadricObj *quadratic = gluNewQuadric();		// Storage For Our Quadratic Object & // Create A Pointer To The Quadric Object
	gluQuadricNormals(quadratic, GLU_SMOOTH);		// Create Smooth Normals

	glPushMatrix();

	glEnable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);

	//glLightModelfv(GL_LIGHT_MODEL_AMBIENT, l->diffuse);
	//glColor4fv(l->diffuse);

	glTranslatef(l->pos.x, l->pos.y, l->pos.z);

	// rotate the objects to point in the right direction
	Vec3D rot(l->pos.x, l->pos.y, l->pos.z);
	//float theta = rot.thetaXZ(l->target);
	//glRotatef(theta * rad2deg, 0.0f, 1.0f, 0.0f);
	
	if (l->type == LIGHT_DIRECTIONAL) { // Directional light
		gluCylinder(quadratic, 0.1, 0.1, 0.7, 6, 6);
	} else if (l->type == LIGHT_POSITIONAL) {	// Positional Light
		gluSphere(quadratic, 0.25, 9, 9);
	} else {	// Spot light
		gluCylinder(quadratic, (l->arc / 360.0f), 0, 0.5, 6, 6);	// A Cone With A Bottom Radius Of .5 And A Height Of 2
	}

	glEnable(GL_LIGHTING);
	glPopMatrix();
}

inline void ModelCanvas::Render()
{
	GLuint lightID;

	glClearColor(bgcolor.x, bgcolor.y, bgcolor.z, 0.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// Test to disable all the excess stuff to try and find the cause of graphical display problems
//	if (!disableHWAcc) {
		// ****************** BACKGROUND IMAGE *******************
		// Draw the background image if any
		if(drawBackground && !drawMask)
		{
			glPushMatrix();

			glMatrixMode(GL_PROJECTION);
			glLoadIdentity();

			glMatrixMode(GL_MODELVIEW);					// Select The Modelview Matrix
			glLoadIdentity();

			glOrtho(0, gXres, 0, gYres, -1.0, 1.0);
			glEnable(GL_TEXTURE_2D);					// Enable 2D Texture Mapping
			glDisable(GL_DEPTH_TEST);
			glDisable(GL_LIGHTING);
			
			glBindTexture(GL_TEXTURE_2D, bgTexture);
			glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
			glBegin(GL_QUADS);
				glTexCoord2f(0.0f, 0.0f); glVertex2i(0, 0);
				glTexCoord2f(1.0f, 0.0f); glVertex2i(gXres, 0);
				glTexCoord2f(1.0f, 1.0f); glVertex2i(gXres, gYres);
				glTexCoord2f(0.0f, 1.0f); glVertex2i(0,  gYres);
			glEnd();

			glPopMatrix();								// Restore The Old Modelview Matrix							
		}
		// ====================================================

		SetupProjection(modelsize);
		
		//if (root->model) {
			// ************* Lighting *******************
			// All our lighting related rendering code
			// Use model lighting?
			if (model && (lightType==LT_MODEL_ONLY)) {
				Vec4D la;

				if (model->header.nLights > 0) {
					la = Vec4D(0.0f, 0.0f, 0.0f, 1.0f);
				} else {
					la = Vec4D(1.0f, 1.0f, 1.0f, 1.0f);
				}

				// Set the Model Ambience lighting.
				glLightModelfv(GL_LIGHT_MODEL_AMBIENT, la);

			} else if (lightType==LT_DIRECTIONAL) {
				for (int i=0; i<MAX_LIGHTS; i++) {
					if (lights[i].enabled && !lights[i].relative) {
						glLightfv(GL_LIGHT0 + i, GL_POSITION, lights[i].pos);
					}
				}
				
			// Ambient lighting is just a single colour applied to all rendered vertices.
			} else if (lightType==LT_AMBIENT) {
				glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lights[0].diffuse);	// use diffuse, as thats our main 'colour setter'
			}
			// ==============================================

			// ************** SKYBOX *************
			// render the skybox, if any
			if (drawSky && skyModel && sky->model) {
				glPushMatrix();
				glLoadIdentity();
				
				float fScale = 64.0f / skyModel->rad;
				
				glTranslatef(0.0f, 0.0f, -5.0f);
				glScalef(fScale, fScale, fScale);
				sky->draw(this);

				glPopMatrix();
			}
			// ====================================

			// *************************
			// setup the view/projection
			if (model && useCamera && model->hasCamera) {
				model->cam.setup();
			} else {
				glTranslatef(vPos.x, vPos.y, -vPos.z);
				glRotatef(vRot.x, 1.0f, 0.0f, 0.0f);
				glRotatef(vRot.y, 0.0f, 1.0f, 0.0f);
				glRotatef(vRot.z, 0.0f, 0.0f, 1.0f);

				// manual fix for furbolgs.
				if (model && !wmo && model->name == "Creature\\Furbolg\\furbolg.m2")
					glRotatef(90.0f, 0.0f, 0.0f, -1.0f);
			}
			// ==========================
			
			// ************* Relative Lighting *******************
			// More lighting code, this is to setup the lights that are 'relative' to the model.
			if (model && (lightType==LT_DIRECTIONAL)) { // Else, for all our models, we use the new "lighting control", IF we're not using model only lighting		
				// loop through the lights of our lighting system checking to see if they are turned on
				// and if so to apply their settings.
				for (int i=0; i<MAX_LIGHTS; i++) {
					if (lights[i].enabled && lights[i].relative) {
						lightID = GL_LIGHT0 + i;

						glLightfv(lightID, GL_POSITION, lights[i].pos);
						
						// Draw our 'light cone' to represent the light.
						if (drawLightDir)
							RenderLight(&lights[i]);
					}
				}
			}
			// ==============================================
			
			// ************* Setup our render state *********
			//glEnable(GL_COLOR_MATERIAL);
			if (drawMask) {
				glDisable(GL_TEXTURE_2D);
				glDisable(GL_LIGHTING);
				//glDisable(GL_DEPTH_TEST);
			} else {
				glEnable(GL_LIGHTING);
				glEnable(GL_TEXTURE_2D);
				//glEnable(GL_DEPTH_TEST);
			}
			// ===============================================
//	}

		// ***************** MODEL RENDERING **********************
		// render our main model
		if (model) {
			//model->animcalc = false;
			glDepthFunc(GL_LEQUAL);
			glEnable(GL_DEPTH_TEST);
			//glDisable(GL_CULL_FACE);
			
			glEnable(GL_NORMALIZE);
			root->draw(this);
			glDisable(GL_NORMALIZE);			
			
			// render our particles, we do this afterwards so that all the particles display "OK" without having things like shields "overwriting" the particles.
			if (!drawMask) {
				glEnable(GL_TEXTURE_2D);
				glDisable(GL_LIGHTING);
				glDepthMask(GL_FALSE);
				//glEnable(GL_ALPHA_TEST);
				glEnable(GL_BLEND);
				
				root->drawParticles();
				
				glDepthMask(GL_TRUE);
				glDisable(GL_ALPHA_TEST);
				glDisable(GL_BLEND);
			}
			
		}
		// ========================================		
		
	//}
	
	//glFlush();
	//glFinish();
	SwapBuffers();
}

inline void ModelCanvas::RenderWMO()
{
	if (!init)
		InitGL();

	glClearColor(bgcolor.x, bgcolor.y, bgcolor.z, 0.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	SetupProjection(modelsize);

	// Lighting
	Vec4D la;
	// From what I can tell, WoW OpenGL only uses 4 lights
	for (int i=0; i<4; i++) {
		GLuint light = GL_LIGHT0 + i;
		glLightf(light, GL_CONSTANT_ATTENUATION, 0.0f);
		glLightf(light, GL_LINEAR_ATTENUATION, 0.7f);
		glLightf(light, GL_QUADRATIC_ATTENUATION, 0.03f);
		glDisable(light);
	}
	la = Vec4D(0.35f, 0.35f, 0.35f, 1.0f);

	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, la);
	glColor3f(1.0f, 1.0f, 1.0f);
	// --==--

	// View
	glTranslatef(vPos.x, vPos.y, -vPos.z);
	glRotatef(vRot.x, 1.0f, 0.0f, 0.0f);
	glRotatef(vRot.y, 0.0f, 1.0f, 0.0f);
	glRotatef(vRot.z, 0.0f, 0.0f, 1.0f);
	//glTranslatef(baseTrans.x, baseTrans.y, baseTrans.z);
	// --==--

	glEnable(GL_DEPTH_TEST);
	glDisable(GL_CULL_FACE);
	root->draw(this);
	//root->drawParticles(true);

	glFlush();
	glFinish();
	SwapBuffers();
}

inline void ModelCanvas::RenderWMOToBuffer()
{
	if (!rt)
		return;

	if (!init || supportFBO)
		InitGL();

	glClearColor(bgcolor.x, bgcolor.y, bgcolor.z, 0.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	if (supportFBO && supportPBuffer) {
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluPerspective( 45.0f, rt->nWidth / rt->nHeight, 3.0f, 1500.0f );

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
	}

	// Lighting
	Vec4D la;
	// From what I can tell, WoW OpenGL only uses 4 lights
	for (int i=0; i<4; i++) {
		GLuint light = GL_LIGHT0 + i;
		glLightf(light, GL_CONSTANT_ATTENUATION, 0.0f);
		glLightf(light, GL_LINEAR_ATTENUATION, 0.7f);
		glLightf(light, GL_QUADRATIC_ATTENUATION, 0.03f);
		glDisable(light);
	}
	la = Vec4D(0.35f, 0.35f, 0.35f, 1.0f);

	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, la);
	glColor3f(1.0f, 1.0f, 1.0f);
	// --==--

	// View
	glTranslatef(vPos.x, vPos.y, -vPos.z);
	glRotatef(vRot.x, 1.0f, 0.0f, 0.0f);
	glRotatef(vRot.y, 0.0f, 1.0f, 0.0f);
	glRotatef(vRot.z, 0.0f, 0.0f, 1.0f);
	//glTranslatef(baseTrans.x, baseTrans.y, baseTrans.z);
	// --==--

	glEnable(GL_DEPTH_TEST);
	glDisable(GL_CULL_FACE);
	root->draw(this);
	//root->drawParticles(true);
	
}

inline void ModelCanvas::RenderToBuffer()
{
	if (!rt)
		return;

	if (!init)
		InitGL();

    if (supportFBO) {
		rt->InitGL();
		glViewport(0,0,rt->nWidth, rt->nHeight);
		SetupProjection();
	}

	GLuint lightID;

	//glViewport( 0, 0, rt->nWidth, rt->nHeight );
	glClearColor(bgcolor.x, bgcolor.y, bgcolor.z, 0.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	// ****************** BACKGROUND IMAGE *******************
	// Draw the background image if any
	if(drawBackground && !drawMask)
	{
		glPushMatrix();

		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();

		glMatrixMode(GL_MODELVIEW);					// Select The Modelview Matrix
		glLoadIdentity();

		glOrtho(0, gXres, 0, gYres, -1.0, 1.0);
		glEnable(GL_TEXTURE_2D);					// Enable 2D Texture Mapping
		glDisable(GL_DEPTH_TEST);
		glDisable(GL_LIGHTING);
		
		glBindTexture(GL_TEXTURE_2D, bgTexture);
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
		glBegin(GL_QUADS);
			glTexCoord2f(0.0f, 0.0f); glVertex2i(0, 0);
			glTexCoord2f(1.0f, 0.0f); glVertex2i(gXres, 0);
			glTexCoord2f(1.0f, 1.0f); glVertex2i(gXres, gYres);
			glTexCoord2f(0.0f, 1.0f); glVertex2i(0,  gYres);
		glEnd();

		glPopMatrix();								// Restore The Old Modelview Matrix							
	}
	// ====================================================

	if (!supportFBO && supportPBuffer) {
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluPerspective( 45.0f, rt->nWidth / rt->nHeight, 0.1f, 100.0f );

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
	}

	lightControl->UpdateGL();
	
	// Use model lighting?
	if (model && (lightType==LT_MODEL_ONLY)) {
		Vec4D la;

		if (model->header.nLights > 0) {
			la = Vec4D(0.0f, 0.0f, 0.0f, 1.0f);
		} else {
			la = Vec4D(1.0f, 1.0f, 1.0f, 1.0f);
		}

		// Set the Model Ambience lighting.
		glLightModelfv(GL_LIGHT_MODEL_AMBIENT, la);

	} else if (lightType==LT_DIRECTIONAL) {
		for (int i=0; i<MAX_LIGHTS; i++) {
			if (lights[i].enabled && !lights[i].relative) {
				lightID = GL_LIGHT0 + i;

				glLightfv(lightID, GL_POSITION, lights[i].pos);
			}
		}
		
	// Ambient lighting is just a single colour applied to all rendered vertices.
	} else if (lightType==LT_AMBIENT) {
		glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lights[0].diffuse);	// use diffuse, as thats our main 'colour setter'
	}

	// ************** SKYBOX *************
	// render the skybox, if any
	if (drawSky && skyModel && sky->model) {
		glPushMatrix();
		glLoadIdentity();
		
		float fScale = 64.0f / skyModel->rad;
		
		glTranslatef(0.0f, 0.0f, -5.0f);
		glScalef(fScale, fScale, fScale);
		sky->draw(this);

		glPopMatrix();
	}
	// ====================================

	// *************************
	// setup the view/projection
	if (model && useCamera && model->hasCamera) {
		model->cam.setup();
	} else {
		glTranslatef(vPos.x, vPos.y, -vPos.z);
		glRotatef(vRot.x, 1.0f, 0.0f, 0.0f);
		glRotatef(vRot.y, 0.0f, 1.0f, 0.0f);
		glRotatef(vRot.z, 0.0f, 0.0f, 1.0f);

		// manual fix for furbolgs.
		if (model && model->name == "Creature\\Furbolg\\furbolg.m2")
			glRotatef(90.0f, 0.0f, 0.0f, -1.0f);

		//glTranslatef(baseTrans.x, baseTrans.y, baseTrans.z);
	}
	// ==========================
		
	// ************* Relative Lighting *******************
	// More lighting code, this is to setup the lights that are 'relative' to the model.
	if (model && (lightType==LT_DIRECTIONAL)) { // Else, for all our models, we use the new "lighting control", IF we're not using model only lighting		
		// loop through the lights of our lighting system checking to see if they are turned on
		// and if so to apply their settings.
		for (int i=0; i<MAX_LIGHTS; i++) {
			if (lights[i].enabled && lights[i].relative) {
				lightID = GL_LIGHT0 + i;

				glLightfv(lightID, GL_POSITION, lights[i].pos);

				// Draw our 'light cone' to represent the light.
				if (drawLightDir)
					RenderLight(&lights[i]);
			}
		}
	}
	// ==============================================
		
	// ************* Setup our render state *********
	//glEnable(GL_COLOR_MATERIAL);
	if (drawMask) {
		glDisable(GL_TEXTURE_2D);
		glDisable(GL_LIGHTING);
		//glDisable(GL_DEPTH_TEST);
	} else {
		glEnable(GL_LIGHTING);
		glEnable(GL_TEXTURE_2D);
		//glEnable(GL_DEPTH_TEST);
	}
	// ===============================================

	// ***************** MODEL RENDERING **********************
	// render our main model
	if (model) {
		glEnable(GL_DEPTH_TEST);
		//glDisable(GL_CULL_FACE);
		
		glEnable(GL_NORMALIZE);
		root->draw(this);
		glDisable(GL_NORMALIZE);			
		
		// render our particles, we do this afterwards so that all the particles display "OK" without having things like shields "overwriting" the particles.
		if (!drawMask) {
			glEnable(GL_TEXTURE_2D);
			glDisable(GL_LIGHTING);
			glDepthMask(GL_FALSE);
			//glEnable(GL_ALPHA_TEST);
			glEnable(GL_BLEND);
			
			root->drawParticles();
			
			glDepthMask(GL_TRUE);
			glDisable(GL_ALPHA_TEST);
			glDisable(GL_BLEND);
		}
		
	}
	// ========================================		
}

void Attachment::draw(ModelCanvas *c)
{
	if (!c)
		return;

	glPushMatrix();
	
	if (model) {
		//if (!parent)
			//glLoadIdentity();

		model->reset();
		setup();

		Model *m = static_cast<Model*>(model);
		
		if (!m)
			return;


		if (!c->wmo) {
			// no need to scale if its already 100%
			if (scale != 1.0f)
				glScalef(scale, scale, scale);

			if (pos != Vec3D(0.0f, 0.0f, 0.0f))
				glTranslatef(pos.x, pos.y, pos.z);

			if (rot != Vec3D(0.0f,0.0f,0.0f)) {
				glRotatef(rot.x, 1.0f, 0.0f, 0.0f);
				glRotatef(rot.y, 0.0f, 1.0f, 0.0f);
				glRotatef(rot.z, 0.0f, 0.0f, 1.0f);
			}

			
			if (m->showModel && (m->alpha!=1.0f)) {
				glDisable(GL_COLOR_MATERIAL);

				float a[] = {1.0f, 1.0f, 1.0f, m->alpha};
				glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, a);

				glEnable(GL_BLEND);
				//glDisable(GL_DEPTH_TEST);
				//glDepthMask(GL_FALSE);
				glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			}

			if (!m->showTexture || c->drawMask)
				glDisable(GL_TEXTURE_2D);
			else
				glEnable(GL_TEXTURE_2D);
		}

		// We call this no matter what so that the model will still 'animate'.
		// and we do the 'showmodel' check inside the function
		model->draw();

		if (!c->wmo) {	
			if (m->showModel && (m->alpha!=1.0f)) {
				float a[] = {1.0f, 1.0f, 1.0f, 1.0f};
				glMaterialfv(GL_FRONT, GL_DIFFUSE, a);

				glDisable(GL_BLEND);
				//glEnable(GL_DEPTH_TEST);
				//glDepthMask(GL_TRUE);
				glEnable(GL_COLOR_MATERIAL);
			}

			// Disable lighting and textures if we haven't already
			if(!c->drawMask) {
				glDisable(GL_LIGHTING);
				glDisable(GL_TEXTURE_2D);
			}

			if (m->showBounds) 
				m->drawBoundingVolume();
				
			if (m->showBones) 
				m->drawBones();
				
			glEnable(GL_LIGHTING);
		}
	}

	// children:
	for (size_t i=0; i<children.size(); i++)
		children[i]->draw(c);
	
	glPopMatrix();
}

void Attachment::drawParticles(bool force)
{
	glPushMatrix();

	if (model) {
		Model *m = static_cast<Model*>(model);
		if (!m)
			return;
		
		//model->reset();
		setupParticle();

		// no need to scale if its already 100%
		if (scale != 1.0f)
			glScalef(scale, scale, scale);

		if (rot != Vec3D(0.0f, 0.0f, 0.0f))
			glRotatef(rot.y, 0.0f, 1.0f, 0.0f);

		//glRotatef(45.0f, 1,0,0);

		if (pos != Vec3D(0.0f, 0.0f, 0.0f))
			glTranslatef(pos.x, pos.y, pos.z);

		if (force)
			m->drawParticles();
		else if (m->hasParticles && m->showParticles) 
			m->drawParticles();
				
	}
	
	// children:
	for (size_t i=0; i<children.size(); i++)
		children[i]->drawParticles();

	glPopMatrix();
}



void Attachment::tick(float dt)
{
	if (model)
		model->update(dt);
	for (size_t i=0; i<children.size(); i++)
		children[i]->tick(dt);
}


void ModelCanvas::OnTimer(wxTimerEvent& event)
{
	if (render) {
		CheckMovement();
		tick();
		Refresh(false);
	}
}

void ModelCanvas::tick()
{
	int ddt = 0;

	// Time stuff
	//time = float();
	ddt = (timeGetTime() - lastTime);// * animSpeed;
	lastTime = timeGetTime();
	// --

	globalTime += (ddt);

	if (model) {
		if (model->animManager && !wmo) {
			if (model->animManager->IsPaused())
				ddt = 0;
			
			if (!model->animManager->IsParticlePaused())
				ddt = model->animManager->GetTimeDiff();
		}
		
		root->tick(ddt);
	}

	if (drawSky && sky && skyModel) {
		sky->tick(ddt);
	}

	
}

/*
void ModelCanvas::TogglePause()
{
	if (!bPaused) {
		// pause
		bPaused = true;
		pauseTime = timeGetTime();

	} else {
		// unpause
		DWORD t = timeGetTime();
		deltaTime += t - pauseTime;
		if (time==0) deltaTime = t;
		bPaused = false;
	}
}
*/

void ModelCanvas::ResetView()
{
	vRot = Vec3D(0,-90,0);
	vPos = Vec3D(0, 0, 5);

	bool isSkyBox = (model->name.substr(0,3)=="Env");
	if (!isSkyBox) {
		if (model->name.find("SkyBox")<model->name.length())
			isSkyBox = true;
	}

	if (isSkyBox) {
		// for skyboxes, don't zoom out ;)
		vPos.y = vPos.z = 0.0f;
	} else {
		vPos.z = model->rad * 1.6f;
		if (vPos.z < 3.0f) vPos.z = 3.0f;
		if (vPos.z > 64.0f) vPos.z = 64.0f;
		
		//ofsy = (model->anims[model->currentAnim].boxA.y + model->anims[model->currentAnim].boxB.y) * 0.5f;
		vPos.y = -model->rad * 0.5f;
		if (vPos.y > 50) vPos.y = 50;
		if (vPos.y < -50) vPos.y = -50;
	}

	modelsize = model->rad * 2.0f;
	
	if (model->name.substr(0,4)=="Item") 
		vRot.y = 0; // items look better facing right by default
}

void ModelCanvas::ResetViewWMO(int id)
{
	if (!wmo || id>=wmo->nGroups) return;

	vRot = Vec3D(0.0f, -90.0f, 0.0f);
	vPos = Vec3D(0.0f, 0.0f, 5.0f);
	Vec3D mid;

	if (id==-1) {
		vPos.z = (wmo->v2-wmo->v1).length();
		mid = (wmo->v1+wmo->v2)*0.3f;
	} else {
		// zoom/center on current WMO group
		WMOGroup &g = wmo->groups[id];
		vPos.z = (g.v2-g.v1).length();
		mid = (g.v1+g.v2)*0.5f;
	}

	modelsize = vPos.z;

	baseTrans = Vec3D(mid.x,mid.z,-mid.y) * -1.0f;

	if (vPos.z < 3.0f) vPos.z = 3.0f;
	if (vPos.z > 500.0f) vPos.z = 500.0f;

	vPos.y = (vPos.y / 2);
}

void ModelCanvas::loadBackground(wxString filename)
{
	FILE *file = NULL;
	BYTE *buffer = NULL;
	//GLint format = GL_RGBA;
	CxImage *image = NULL;

	if (!wxFile::Exists(filename))
		return;

	file = fopen(filename.fn_str(), "rb");
	if (!file) {
		fclose(file);
		return;
	}

	bgImagePath = filename;

	// Setup the OpenGL Texture stuff
	glGenTextures(1, &bgTexture);
	glBindTexture(GL_TEXTURE_2D, bgTexture);
	
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);	// Linear Filtering
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);	// Linear Filtering

	// Get the file extension and load the file
	wxString tmp = filename.Mid(filename.Length() - 3, 3);
	tmp.LowerCase();

	if (tmp == "bmp")
		image = new CxImage(file, CXIMAGE_FORMAT_BMP);
	else if (tmp == "tga")
		image = new CxImage(file, CXIMAGE_FORMAT_TGA);
	else if (tmp == "jpg")
		image = new CxImage(file, CXIMAGE_FORMAT_JPG);
	else {
		fclose(file);
		return;
	}

	if (image == NULL) {
		fclose(file);
		return;
	}

	// TODO: Check the size of the images, resize as needed?
	bool textureXSafe = false;
	bool textureYSafe = false;
	for (unsigned int i=2; i < 2100; i+=i) {
		if (image->GetWidth() == i)
			textureXSafe = true;
		if (image->GetHeight() == i)
			textureYSafe = true;
	}

	if (!textureXSafe || !textureYSafe) {
		wxMessageBox(_T("The dimensions of that image are not to a power of 2.  eg. 64x64, 128x128, 256x256, 512x512\nThis may result in unstable results, resizing to 256x256"), _T("Caution!"));
		if (image->GetBpp() < 24)
			image->IncreaseBpp(24);
		image->Resample(256,256);
	}

	long size = image->GetWidth() * image->GetHeight() * 4;
	image->Encode2RGBA(buffer, size);
	
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, image->GetWidth(), image->GetHeight(), 0, GL_RGBA, GL_UNSIGNED_BYTE, buffer);
	drawBackground = true;

	fclose(file);

	//delete file;
	wxDELETE(image);
	wxDELETE(buffer);
}

void ModelCanvas::Zoom(float f, bool rel)
{
	if (rel) {
		float cosx = cos(vRot.x * piover180);
		vPos.x += cos(vRot.y * piover180) * cosx * f;
		vPos.y += sin(vRot.x * piover180) * sin(vRot.y * piover180) * f;
		vPos.z += sin(vRot.y * piover180) * cosx * f;
	} else {
		vPos.z -= f;
	}

	viewControl->Refresh();
}

// Check for keyboard input
void ModelCanvas::OnKey(wxKeyEvent &event)
{
	// error checking
	if(!model) 
		return;
	
	// --
	int keycode = event.GetKeyCode(); 

	// if(bInGameKeys) {
		
	// } else {
		// animation speed toggles
		if (keycode == '0')	animControl->SetAnimSpeed(1.0f);
		else if (keycode == '1')	animControl->SetAnimSpeed(0.1f);
		else if (keycode == '2')	animControl->SetAnimSpeed(0.2f);
		else if (keycode == '3')	animControl->SetAnimSpeed(0.3f);
		else if (keycode == '4')	animControl->SetAnimSpeed(0.4f);
		else if (keycode == '5')	animControl->SetAnimSpeed(0.5f);
		else if (keycode == '6')	animControl->SetAnimSpeed(0.6f);
		else if (keycode == '7')	animControl->SetAnimSpeed(0.7f);
		else if (keycode == '8')	animControl->SetAnimSpeed(0.8f);
		else if (keycode == '9')	animControl->SetAnimSpeed(0.9f);
		// --	
	// }
}

void ModelCanvas::CheckMovement()
{
	if (!model || wmo || !model->animManager)
		return;

	// Make sure its the canvas that has focus before continuing
	wxWindow *win = wxWindow::FindFocus();
	if(!win)
		return;
	wxGLCanvas *gl = wxDynamicCast(win, wxGLCanvas);
	if(!gl)
		return;

	float speed;

	// Time stuff
	if (!wmo)
		speed = ((timeGetTime() - lastTime) * model->animManager->GetSpeed()) / 7.0f;
	else 
		speed = (timeGetTime() - lastTime);

	//lastTime = timeGetTime();

	// Turning
	if (wxGetKeyState(WXK_LEFT)) {
		vRot.y += speed;
		viewControl->Refresh();

		if (vRot.y > 360) vRot.y -= 360;
		if (vRot.y < 0) vRot.y += 360;

	} else if (wxGetKeyState(WXK_RIGHT)) {
		vRot.y -= speed;
		viewControl->Refresh();

		if (vRot.y > 360) vRot.y -= 360;
		if (vRot.y < 0) vRot.y += 360;
	}
	// --

	// Moving forward/backward
	//float speed = 0.0f;
	if (model->animated)
		speed *= (model->anims[model->currentAnim].moveSpeed / 160.0f);
	//else
	//	speed *= 0.05f;

	if (wxGetKeyState(WXK_UP))
		Zoom(speed, true);
	else if (wxGetKeyState(WXK_DOWN))
		Zoom(-speed, true);
	// --
}

// Our screenshot function which supports both PBO and FBO aswell as traditional older cards.
void ModelCanvas::Screenshot(const wxString fn, int x, int y)
{
	if (!init)
		InitGL();

	glPixelStorei(GL_PACK_ALIGNMENT, 1);

	wxFileName temp(fn, wxPATH_NATIVE);
	
	unsigned char *pixels = NULL;
	int screenSize[4];

	// Setup out buffers for offscreen rendering
	if (supportPBuffer || supportFBO) {
		rt = new RenderTexture();
		rt->Init((HWND)this->GetHandle(), x, y, supportFBO);
	}


	//glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	if (supportPBuffer || supportFBO) { // if either are supported use our 'RenderTexture' object.
		screenSize[2] = rt->nWidth;
		screenSize[3] = rt->nHeight;
		rt->BeginRender();
		//rt->BindTexture(); 

		if (wmo)
			RenderWMOToBuffer();
		else
			RenderToBuffer();

	} else {		
		glGetIntegerv(GL_VIEWPORT, screenSize);
	}
	
	// (width*height*bytesPerPixel) - 32bit, 4 bytes
	pixels = new unsigned char[screenSize[2]*screenSize[3]*4];

	// read in the pixel data
	glReadPixels(0, 0, screenSize[2], screenSize[3], GL_BGRA_EXT, GL_UNSIGNED_BYTE, pixels);
	
	if (supportPBuffer || supportFBO) {
		//rt->UnBindTexture();
		rt->EndRender();
	}

	CxImage *newImage = new CxImage(0);
	newImage->AlphaCreate();
	newImage->IncreaseBpp(32);
	newImage->CreateFromArray(pixels, screenSize[2], screenSize[3], 32, (screenSize[2]*4), false);
	
	bool succ = true;

	// Save
	if (temp.GetExt() == "tga") {
		newImage->Save(fn.fn_str(), CXIMAGE_FORMAT_TGA);
	//} else if (temp.GetExt() == "tif") {
	//	newImage->Save(fn.fn_str(), CXIMAGE_FORMAT_TIF);
	} else if (temp.GetExt() == "png") {
		newImage->Save(fn.fn_str(), CXIMAGE_FORMAT_PNG);
	} else if (temp.GetExt() == "jp2") {
		newImage->Save(fn.fn_str(), CXIMAGE_FORMAT_JP2);
	} else if (temp.GetExt() == "jpg") {
		newImage->SetJpegQuality(100);
		newImage->SetJpegScale(100);
		newImage->Save(fn.fn_str(), CXIMAGE_FORMAT_JPG);
	} else if (temp.GetExt() == "bmp") // Save Bitmap format
		newImage->Save(fn.fn_str(), CXIMAGE_FORMAT_BMP);
	else
		succ = false;

	if (succ)
		wxLogMessage(_T("Screenshot saved to: %s"), fn.c_str());

	newImage->Destroy();
	wxDELETE(newImage);

	// free the memory
	wxDELETEA(pixels);

	if (supportPBuffer || supportFBO) {
		rt->Shutdown();
		wxDELETE(rt);
	}
}

void Attachment::setup()
{
	if (parent==0) 
		return;
	if (parent->model) 
		parent->model->setupAtt(id);
}

void Attachment::setupParticle()
{
	if (parent==0) 
		return;
	if (parent->model) 
		parent->model->setupAtt2(id);
}

Attachment* Attachment::addChild(const char *modelfn, int id, int slot, float scale, float rot, Vec3D pos)
{
	if (!modelfn || !strlen(modelfn) || id<0) 
		return 0;

	Model *m = new Model(modelfn, true);

	if (m && m->ok) {
		return addChild(m, id, slot, scale, rot, pos);
	} else {
		wxDELETE(m);
		return 0;
	}
}

Attachment* Attachment::addChild(Displayable *disp, int id, int slot, float scale, float rot, Vec3D pos)
{
	Attachment *att = new Attachment(this, disp, id, slot, scale, rot, pos);
	children.push_back(att);
	return att;
}

void Attachment::delChildren()
{
	for (size_t i=0; i<children.size(); i++) {
		children[i]->delChildren();

		wxDELETE(children[i]->model);
		wxDELETE(children[i]);
	}

	children.clear();
}

void Attachment::delSlot(int slot)
{
	for (size_t i=0; i<children.size(); ) {
		if (children[i]->slot == slot) {
			wxDELETE(children[i]);
			children.erase(children.begin() + i);
		} else i++;
	}
}

